@@ -2,7 +2,8 @@ module TestBlockIndices
 
 using BlockArrays, FillArrays, Test, StaticArrays, ArrayLayouts
 using OffsetArrays
-import BlockArrays: BlockIndex, BlockIndexRange, BlockSlice, BlockedSlice
+import BlockArrays: BlockIndex, BlockIndexRange, BlockSlice, NoncontiguousBlockSlice
+import BlockArrays: split_index, merge_indices
 
 @testset "Blocks" begin
     @test Int(Block(2)) === Integer(Block(2)) === Number(Block(2)) === 2
@@ -86,11 +87,21 @@ import BlockArrays: BlockIndex, BlockIndexRange, BlockSlice, BlockedSlice
     @testset "BlockIndex" begin
         @test Block()[] == BlockIndex()
         @test Block(1)[1] == BlockIndex((1,),(1,))
+        @test Block(1)[Block(1)] == BlockIndex((1,),(Block(1),))
         @test Block(1)[1:2] == BlockIndexRange(Block(1),(1:2,))
+        @test Block(1)[[1,3]] == BlockIndices(Block(1),([1,3],))
         @test Block(1,1)[1,1] == BlockIndex((1,1),(1,1)) == BlockIndex((1,1),(1,))
         @test Block(1,1)[1:2,1:2] == BlockIndexRange(Block(1,1),(1:2,1:2))
+        @test Block(1,1)[[1,3],1:2] == BlockIndices(Block(1,1),([1,3],1:2))
         @test BlockIndexRange((Block(1)[1:2],Block(1)[1:2])) == BlockIndexRange(Block(1,1),(1:2,1:2))
+        @test BlockIndices((Block(1)[[1,3]],Block(1)[[2,4]])) == BlockIndices(Block(1,1),([1,3],[2,4]))
         @test Block(1)[1:3][1:2] == BlockIndexRange(Block(1),1:2)
+        @test Block(1)[[1,3,5]][[1,3]] == BlockIndices(Block(1),[1,5])
+        @test Block(1,2)[[1,3,5],[2,4,6]][2,3] == BlockIndex(Block(1,2),(3,6))
+        @test Block(1)[[1,3,5]][2:3] == BlockIndices(Block(1),[3,5])
+        @test Block(1)[2:4][[1,3]] == BlockIndices(Block(1),[2,4])
+        @test Block(1,1)[1:3,1:3][1:2,1:2] == BlockIndexRange(Block(1,1),1:2,1:2)
+        @test Block(1,1)[1:3,1:3][1:2,[1,3]] == BlockIndices(Block(1,1),1:2,[1,3])
         @test Block(1,1)[2:4,2:4][2:3,2:3] == BlockIndexRange(Block(1,1),(3:4,3:4))
         @test BlockIndexRange(Block(),())[] == BlockIndex()
         @test BlockIndex((2,2,2),(2,)) == BlockIndex((2,2,2),(2,1,)) == BlockIndex((2,2,2),(2,1,1))
@@ -98,7 +109,16 @@ import BlockArrays: BlockIndex, BlockIndexRange, BlockSlice, BlockedSlice
         @test BlockIndex(UInt(2),(2,)) === BlockIndex((UInt(2),),(2,))
         @test BlockIndex(Block(2),2) === BlockIndex(Block(2),(2,))
         @test BlockIndex(Block(2),UInt(2)) === BlockIndex(Block(2),(UInt(2),))
+        @test BlockIndex(Block(2),Block(2)) === BlockIndex(Block(2),(Block(2),))
         @test copy(Block(1)[1:2]) === BlockIndexRange(Block(1),1:2)
+        @test copy(Block(1)[[1,3]]) == BlockIndices(Block(1),[1,3])
+        @test copy(Block(1)[[1,3]]) ≢ BlockIndices(Block(1),[1,3])
+    end
+
+    @testset "BlockIndices" begin
+        @test eltype(BlockIndices{3}) ≡ BlockIndex{3}
+        @test Base.IteratorSize(BlockIndices{3}) ≡ Base.HasShape{1}()
+        @test isnothing(iterate(Block(3,3)[[1,3],[3,1]]))
     end
 
     @testset "BlockRange" begin
@@ -615,6 +635,7 @@ end
         b = blockedrange([1,2,3])
         @test b[Block(3)[2]] == b[Block(3)][2] == 5
         @test b[Block(3)[2:3]] == b[Block(3)][2:3] == 5:6
+        @test b[Block(3)[[3,2]]] == b[Block(3)][[3,2]] == [6,5]
     end
 
     @testset "BlockRange indexing" begin
@@ -864,6 +885,11 @@ end
     @test b[1:2] ≡ b[1:2][1:2] ≡ BlockSlice(Block(5)[1:2],1:2)
     @test Block(b) ≡ Block(5)
 
+    bi = BlockSlice(Block(2)[2:4],3:5)
+    @test Block(bi) ≡ Block(2)
+    @test bi[2:3] ≡ BlockSlice(Block(2)[3:4],4:5)
+    @test Block(BlockSlice(Block(2)[1:2], 3:4)) ≡ Block(2)
+
     @testset "OneTo converts" begin
         for b in (BlockSlice(Block(1), 1:1), BlockSlice(Block.(1:1), 1:1), BlockSlice(Block(1)[1:1], 1:1))
             @test convert(typeof(b), Base.OneTo(1)) ≡ b
@@ -881,14 +907,24 @@ end
     end
 end
 
-@testset "BlockedSlice" begin
-    b = BlockedSlice([Block(2), Block(1)], mortar([3:5, 1:2]))
-    @test length(b) == 5
-    for i in eachindex(b.indices)
-        @test b[i] === b.indices[i]
+@testset "NoncontiguousBlockSlice" begin
+    bs = NoncontiguousBlockSlice([Block(2),Block(1)], mortar([3:5,1:2]))
+    @test length(bs) ≡ 5
+    for i in eachindex(bs.indices)
+        @test bs[i] ≡ bs.indices[i]
     end
-    @test b[Block(1)] === BlockSlice(Block(2), 3:5)
-    @test b[Block(2)] === BlockSlice(Block(1), 1:2)
+    @test bs[Block(1)] ≡ BlockSlice(Block(2), 3:5)
+    @test bs[Block(2)] ≡ BlockSlice(Block(1), 1:2)
+    @test BlockArrays._indices(bs) == mortar([3:5,1:2])
+
+    b = NoncontiguousBlockSlice(Block(3), 2:4)
+    @test b[2:3] ≡ NoncontiguousBlockSlice(Block(3)[2:3], 3:4)
+    @test Block(b) ≡ Block(3)
+    @test BlockArrays._indices(b) ≡ 2:4
+
+    bir = NoncontiguousBlockSlice(Block(3)[3:5], 4:6)
+    @test Block(bir) ≡ Block(3)
+    @test bir[2:3] ≡ NoncontiguousBlockSlice(Block(3)[4:5], 5:6)
 end
 
 #=
@@ -983,9 +1019,30 @@ end
     @test !blockisapprox(B1, B12; rtol=1e-9)
 end
 
-@testset "BlockIndices" begin
+@testset "Vector{<:Block{1}}" begin
     a = BlockedOneTo(1:3)
     @test a[[Block(1),Block(3)]] == a[Block.(1:2:3)] == [1,3]
 end
 
+@testset "split_index" begin
+    @test split_index(CartesianIndex(1, 2)) ≡ (1, 2)
+    @test split_index(Block(1, 2)) ≡ (Block(1), Block(2))
+    @test split_index(Block(1, 2)[3, 4]) ≡ (Block(1)[3], Block(2)[4])
+    @test split_index(Block(1, 2)[3:4, 4:5]) ≡ (Block(1)[3:4], Block(2)[4:5])
+end
+
+@testset "merge_indices" begin
+    @test merge_indices((1, 2)) ≡ CartesianIndex(1, 2)
+    @test merge_indices((Block(1), Block(2))) ≡ Block(1, 2)
+    @test merge_indices((Block(1)[3], Block(2)[4])) ≡ Block(1, 2)[3, 4]
+    @test merge_indices((Block(1)[3:4], Block(2)[4:5])) ≡ Block(1, 2)[3:4, 4:5]
+end
+@testset "to_index" begin
+    @test_throws ArgumentError Base.to_index(Block(3))
+    @test_throws ArgumentError Base.to_index(Block(3)[2])
+    @test_throws ArgumentError Base.to_index(Block(3)[1:2])
+    @test_throws ArgumentError Base.to_index(Block(3)[[1,3]])
+    @test_throws ArgumentError Base.to_index(Block.(2:3))
+end
+
 end # module
